import numpy as np

import fastash.ashfunc
from .whitening_transformer import WhiteningTransformer
from hyperclip import Hyperplane
from joblib import Parallel, delayed, dump, load


import os
import shutil

class ASH():
    """
    This class implements the averaged shifted histogram method.
    
    Parameters
    ----------
    h : {'scott'} or float, default='scott'
        The bin width value. If ``'scott'``, the Scott's rule is applied.
    
    q : int, default=100
        The number of histograms
    
    bounds : list of (feature_id, type)
        The bounds of the case. For exemple, if it is a 2d case with the second feature low bounded, the parameters should be ``[1, 'left']``. The type can be 'left', 'right' or 'both'.
    
    n_mc : int, default==10**4
        The number of elements of the MonteCarlo process used in the boundaries correction algorithm.
        
    n_jobs : int, default=1
        The number of parallel jobs to run for neighbors search.
        ``None`` means 1 unless in a :obj:`joblib.parallel_backend` context.
        ``-1`` means using all processors.
    
    verbose : int, default=0
        Verbosity level.
    """
    def __init__(self,
                 h='scott',
                 q=100,
                 bounds = [],
                 n_mc = 10**4,
                 n_jobs = 1,
                 verbose=0):

        self.h = h
        self._h = None
        self.q = q
        self.n_mc = n_mc
        self.bounds = bounds 
        self.n_jobs = n_jobs
        self.verbose = verbose

    def __repr__(self):
        if self._h is None:
            return('ASH(h='+str(self.h)+')')
        else:
            return('ASH(h='+str(self._h)+')')

    def fit(self, X):
        """
        fit
        """
        # preprocessing
        if len(X.shape) == 1:
            X = X[:,None]
        
        self._real_X_min = np.min(X, axis=0)
        self._real_X_max = np.max(X, axis=0)
        
        # get data dimensions
        self._n, self._d = X.shape
        
        # preprocessing
        self._wt = WhiteningTransformer()
        X = self._wt.fit_transform(X)
        
        self._x_min = np.min(X, axis=0)
        
        # BOUNDARIES INFORMATIONS
        A, R = self._set_boundaries(x = X[0])

        # BANDWIDTH SELECTION
        self._compute_bandwidth(X)
        
        # MonteCarlo Set
        # used for the volume computation of hypercubes clipped by hyperplans
        # X_mc = np.random.random((self.n_mc, self._d)) * self._h - self._h / 2
                
        folder = './.temp_joblib_memmap'
        try:
            os.mkdir(folder)
        except FileExistsError:
            pass
        
        X_filename_memmap = os.path.join(folder, 'X_memmap')
        dump(X, X_filename_memmap)
        X_memmap = load(X_filename_memmap, mmap_mode='r')

        self._fit_results = Parallel(n_jobs=self.n_jobs)(delayed(fastash.ashfunc.compute_bins_P_shift)(
            X_memmap, # X
            X.min(axis=0), # x_min
            np.linalg.norm(A, axis=0, ord=np.inf), #norm_inf_w
            np.linalg.norm(A, axis=0, ord=2), # norm_2_w
            self._h, # h
            i_shift * self._h / self.q, # shift
            A, # A
            R, # R
            ) for i_shift in range(self.q))
        
        try:
            shutil.rmtree(folder)
        except:  # noqa
            print('Could not clean-up automatically.')
        
        return(self)

    def predict(self, X, y=None):
        """
        predict
        """
        if len(X.shape) == 1:
            X = X[:,None]
        
        X = self._wt.transform(X)
        
        id_out_of_bounds = np.zeros(X.shape[0]).astype(np.bool)
        for hyp in self._bounds_hyperplanes:
            id_out_of_bounds = np.any((id_out_of_bounds, ~hyp.side(X)), axis=0)
                
        folder = './.temp_joblib_memmap'
        try:
            os.mkdir(folder)
        except FileExistsError:
            pass
        
        X_filename_memmap = os.path.join(folder, 'X_memmap')
        dump(X, X_filename_memmap)
        X_memmap = load(X_filename_memmap, mmap_mode='r')
        
        f = Parallel(n_jobs=self.n_jobs)(delayed(fastash.ashfunc.merge_predict_shift)(
            self._fit_results[i_shift][0], # X_digit_uniques
            self._fit_results[i_shift][1], # P
            X_memmap, # Y
            self._x_min, # x_min
            self._h, # h
            self._fit_results[i_shift][2], # shift
            ) for i_shift in range(self.q))
        
        f = np.sum(f, axis=0)
        f = f / (self.q * self._h ** self._d)
        
        # outside bounds : equal to 0
        f[id_out_of_bounds] = 0

        # Preprocessing correction
        f /= self._wt.scale_
        
        try:
            shutil.rmtree(folder)
        except:  # noqa
            print('Could not clean-up automatically.')
        
        return(f)
    
    def _set_boundaries(self, x):
        self._bounds_hyperplanes = []

        for k, pos in self.bounds:
            if pos == 'left':
                self._add_boundary(k=k,
                                   value=self._real_X_min[k],
                                   x=x)
            elif pos == 'right':
                self._add_boundary(k=k,
                                   value=self._real_X_max[k],
                                   x=x)
            elif pos == 'both':
                self._add_boundary(k=k,
                                   value=self._real_X_min[k],
                                   x=x)
                self._add_boundary(k=k,
                                   value=self._real_X_max[k],
                                   x=x)
            else:
                raise(TypeError('Unexpected bounds parameters'))
        
        A = np.zeros((self._d, len(self._bounds_hyperplanes)))
        R = np.zeros(len(self._bounds_hyperplanes))
        
        for i_hyp, hyp in enumerate(self._bounds_hyperplanes):
            A[:, i_hyp] = hyp.a
            R[i_hyp] = hyp.r
        
        return(A, R)
    
    def _add_boundary(self, k, value, x):
        P = np.diag(np.ones(self._d))
        
        P[:, k] = value

        P_wt = self._wt.transform(P)

        hyp = Hyperplane().set_by_points(P_wt)
        hyp.set_positive_side(x)
        self._bounds_hyperplanes.append(hyp)
    
    def _compute_bandwidth(self, X):
        if type(self.h) is int or type(self.h) is float:
            self._h = float(self.h)

        elif type(self.h) is str:
            if self.h == 'scott' or self.h == 'silverman':
                # the scott rule is based on gaussian kernel
                # the support of the gaussian kernel to have 99%
                # of the density is 2.576
                self._h = 2.576 * scotts_rule(X)
            else:
                raise (ValueError("Unexpected bandwidth selection method."))
        else:
            raise (TypeError("Unexpected bandwidth type."))

        if self.verbose > 0:
            print('Bandwidth selection done : h=' + str(self._h))

    def set_params(self, **params):
        """
        Set parameters.

        Parameters
        ----------
        **params : kwargs
            Parameters et values to set.

        Returns
        -------
        self : DensityEstimator
            The self object.

        """
        for param, value in params.items():
            setattr(self, param, value)

def scotts_rule(X):
    """
    Scott's rule according to "Multivariate density estimation", Scott 2015, p.164.
    The Silverman's rule is exactly the same in "Density estimation for statistics and data analysis", Silverman 1986, p.87.
    Parameters
    ----------
    X : numpy array of shape (n_samples, n_features).

    Returns
    -------
    h : float
        Scotts rule bandwidth. The returned bandwidth should be then factored
        by the data variance.

    """
    n = X.shape[0]
    d = X.shape[1]
    return(n**(-1/(d + 4)))


# -*- coding: utf-8 -*-

import numpy as np


class WhiteningTransformer():
    """
    Whitening transformation in order to have a covariance matrix equal to the identity matrix.
    """

    def fit(self, X):
        """
        Fit the transformer

        Parameters
        ----------
        X : array-like of shape (n_samples, n_features)
            Data samples.

        Returns
        -------
        self : _WhiteningTransformer
            The fitted object.

        """

        self._mean = X.mean(axis=0)

        self._num_obs = X.shape[0]

        _, self._s, Vt = np.linalg.svd(X - self._mean, full_matrices=False)
        self._V = Vt.T

        self._transform_matrix = self._V @ np.diag(1 / self._s) * np.sqrt(self._num_obs - 1)
        self._inverse_transform_matrix = np.diag(self._s) @ self._V.T / np.sqrt(self._num_obs - 1)

        self._transform_det = np.abs(np.linalg.det(self._transform_matrix))
        self._inverse_transform_det = np.abs(np.linalg.det(self._inverse_transform_matrix))

        # for compatibility with other preprocessing methods
        self.scale_ = self._inverse_transform_det

        return (self)

    def transform(self, X):
        """
        Transform.

        Parameters
        ----------
        X : array-like of shape (n_samples, n_features)
            Data samples.

        Returns
        -------
        X_wt : ndarray of shape (n_samples, n_features)
            Transformed samples.

        """
        X = X - self._mean
        return (X @ self._transform_matrix)

    def inverse_transform(self, X):
        """
        inverse transform.

        Parameters
        ----------
        X : array-like of shape (n_samples, n_features)
            Data samples in the Whitening Transform space.

        Returns
        -------
        X_inv_wt : ndarray of shape (n_samples, n_features)
            Inverse transformed samples.

        """
        X = X @ self._inverse_transform_matrix
        return (X + self._mean)

    def fit_transform(self, X):
        """
        fit and transform.

        Parameters
        ----------
        X : array-like of shape (n_samples, n_features)
            Data samples.

        Returns
        -------
        X_wt : ndarray of shape (n_samples, n_features)
            Transformed samples.
        """
        self.fit(X)
        return (self.transform(X))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Sep 14 09:39:01 2021

@author: frem
"""

import numpy as np


class Hyperplane():
    """
    Hyperplane object. Used for boundary bias correction within the GKDE method.
    """

    def __init__(self,
                 w=None,
                 b=None,
                 positive_side_scalar=1):
        self.w = w
        self.b = b
        self.positive_side_scalar = positive_side_scalar

    def set_by_points(self, A):
        """
        Set the hyperplane by points.

        Parameters
        ----------
        A : array-like of shape (n_samples, n_features)
            The points which define the hyperplane. Each row is a point.
            For `$d$` dimension, ``A`` should be of shape `$(d,d)$`.

        Returns
        -------
        self : Hyperplane
            The set hyperplane.

        """
        if A.shape[0] != A.shape[1]:
            raise (ValueError('Unexpected A value. For a d dimension problem, it should be of shape (d,d).'))

        self.A = A
        self.w = np.linalg.solve(A, np.ones(A.shape[0]))
        self.b = - np.dot(self.w, A[0])

        return (self)

    def distance(self, X, p=2):
        """
        Computes the distance to the hyperplane according to the formula :

        ..math::
            dist = \frac{x \dot w + b}{\| w \|}

        Parameters
        ----------
        X : array-like of shape (n_samples, n_features)
            The points to compute the distance to the hyperplane.

        Returns
        -------
        dist : array-like of shape (n_samples,).
            The computed distances.


        """
        dist = np.abs(np.dot(X, self.w) + self.b) / np.linalg.norm(self.w, ord=2)**2 * np.linalg.norm(self.w, ord=p)

        return (dist)

    def set_positive_side(self, P):
        norm_P = np.dot(P[None, :], self.w) + self.b

        if norm_P == 0:
            raise (ValueError("P should not belongs to the hyperplane."))
        
        if norm_P < 0:
            self.w *= -1
            self.b *= -1
        # if norm_P > 0:
            # self.positive_side_scalar = 1
        # else:
            # self.positive_side_scalar = -1

    def side(self, X):
        norm_vec = np.dot(X, self.w) + self.b
        norm_vec *= self.positive_side_scalar

        return (norm_vec > 0)
    
    def affine_transform(self, delta, alpha, inplace=False):
        t_w = np.zeros_like(self.w)
        for j in range(self.w.size):
            if ~np.isclose(self.w[j],0):
                t_w[j] = self.w[j] / alpha
                
        t_b = self.b - np.dot(self.w, delta)
        
        if inplace:
            self.w = t_w
            self.b = t_b
            return(self)
        else:
            return(Hyperplane(w = t_w, b = t_b, positive_side_scalar = self.positive_side_scalar))
    
    def compute_n_solutions(self):
        sol = []
        list_j = []
        for j in range(self.w.size):
            if ~np.isclose(self.w[j], 0):
                s = np.zeros_like(self.w)
                s[j] = -self.b / self.w[j]
                sol.append(s)
                list_j.append(j)
        
        list_j_bar = np.delete(np.arange(self.w.size), list_j)
        
        complete_sol = np.zeros((len(self.w), len(self.w)))
        complete_sol[list_j,:] = np.array(sol)
        
        for j in range(self.w.size):
            if j not in list_j:
                complete_sol[j,:] = sol[0]
                complete_sol[list_j_bar, j] = 1 
        
        return(complete_sol)



# -*- coding: utf-8 -*-

from scipy import stats
import numpy as np
from matplotlib import pyplot as plt
import fastash
from time import time

#%% Dataset

X_min = np.array([-0.5, 0.3])
x1_max = 1.5

mean = np.array([0,0])
rho = 0.7
cov = np.array([[1,rho],[rho,1]])

def _pdf(X, X_min):
    f = stats.multivariate_normal.pdf(X, mean=mean, cov=cov) 

    f[np.any(X<X_min, axis=1)] = 0
    f[X[:,1]>=x1_max] = 0

    return(f)

def _rvs(X_min, n,seed=None):
    np.random.seed(seed)
    X = stats.multivariate_normal.rvs(mean=mean, cov=cov, size=n)
    np.random.seed(None)
    X = X[np.all(X >= X_min, axis=1)]
    X = X[X[:,1] < x1_max]

    return(X)

def bounded_set(n, seed):
    X = _rvs(X_min, n, seed=seed)
    
    xk = (np.linspace(X[:,0].min()-X[:,0].std(),X[:,0].max()+X[:,0].std(),300),
          np.linspace(X[:,1].min()-X[:,1].std(),X[:,1].max()+X[:,1].std(),300))
    X_grid = np.meshgrid(xk[0], xk[1])
    X_grid = np.vstack((X_grid[0].flat, X_grid[1].flat)).T
    
    pdf_grid = np.sum(_pdf(X_grid, X_min)) * np.product(X_grid.max(axis=0)-X_grid.min(axis=0)) / X_grid.shape[0]# -*- coding: utf-8 -*-
    
    Y = np.vstack((np.ones(100)*0.5,
                   np.linspace(0,2,100))).T
    
    pdf_Y = _pdf(Y, X_min) / pdf_grid
    
    return(X, Y, pdf_Y, X_grid)
    

# %%
X, Y, pdf_Y, X_grid = bounded_set(100000, 42)

#%%
st = time()
ash = fastash.ASH(q=100,
        bounds=[
                (0, 'left'),
                (1, 'both')
                ],
        n_mc=10000,
        n_jobs=1)
ash.fit(X)
print('fit exec time : ', time()-st)

# st = time()
# n_X = np.max([ash._fit_results[i_shift][0].shape[0] for i_shift in range(ash.q)])

# ash._X_digit_uniques = np.zeros((ash.q, n_X, ash._d))
# ash._P = np.zeros((ash.q, n_X))
# for i_shift in range(ash.q):
#     n_X_i = ash._fit_results[i_shift][0].shape[0]
    
#     ash._X_digit_uniques[i_shift, :n_X_i, :] = ash._fit_results[i_shift][0]
#     ash._P[i_shift, :n_X_i] = ash._fit_results[i_shift][1]

# ash._X_digit_uniques = ash._X_digit_uniques.astype(np.int32)

# print('self process exec time :', time()-st)

# print(len(ash._ret))
# f = ash.predict(X_grid)

# I = np.sum(f) * np.product(X_grid.max(axis=0)-X_grid.min(axis=0)) / X_grid.shape[0]
# print(I)

# st = time()
# f_grid = ash.predict(X_grid)
# print('grid exec time :', time()-st)

#%%
X_eval = np.vstack((np.ones(100)*0.5,
                        np.linspace(0,2,100))).T

st = time()
f_eval = ash.predict(X_eval)
print('eval exec time :', time()-st)

plt.plot(Y[:,1], pdf_Y)
plt.plot(Y[:,1], f_eval)
plt.show()

#%%
from sklearn.neighbors import KernelDensity
kde = KernelDensity(kernel='gaussian', bandwidth=ash._h / 2.576).fit(X)

st = time()
f_eval_kde = np.exp(kde.score_samples(X_eval))
print('eval exec time :', time()-st)

plt.plot(Y[:,1], pdf_Y)
plt.plot(Y[:,1], f_eval)
plt.plot(Y[:,1], f_eval_kde)
plt.show()



# distutils: language = c++

cdef extern from "stdlib.h":
    ctypedef void const_void "const void"
    void qsort(void *base, int nmemb, int size,
            int(*compar)(const void *, const void *)) nogil

from libcpp.vector cimport vector
from libc.stdlib cimport malloc, free
from libcpp cimport bool

from hyperclip import hyperfunc

import cython


import numpy as np
cimport numpy as np

np.import_array()

from cpython cimport array
import array

@cython.boundscheck(False)  # Deactivate bounds checking.
@cython.wraparound(False)   # Deactivate negative indexing.
@cython.cdivision(True) # Deactivate zero division checking.
def compute_bins_P_shift(np.ndarray[double, ndim=2] X, 
                         np.ndarray[double, ndim=1] x_min,
                         np.ndarray[double, ndim=1] norm_inf_w,
                         np.ndarray[double, ndim=1] norm_2_w,
                         double h, 
                         double shift, 
                         np.ndarray[double, ndim=2] A,
                         np.ndarray[double, ndim=1] R):
    cdef Py_ssize_t i, j, j_unique, i_hyp
    
    cdef int n = X.shape[0]
    cdef int d = X.shape[1]
    
    cdef int n_hyp = A.shape[1]
        
    X_digit_uniques = np.zeros((n, d), dtype=np.intc)
    cdef np.ndarray[int, ndim=2] X_digit_uniques_view = X_digit_uniques
    
    cdef double [:, :] A_x
    cdef double [:] R_x
        
    P = np.zeros(n, dtype=np.double)
    cdef np.ndarray[double, ndim=1] P_view = P
    
    cdef int **X_digit
            
    cdef bool trigger_unique, trigger_correction
    
    cdef int * last_x
    last_x = <int *> malloc(d * sizeof(int))
    
    cdef int i_last = 0
    cdef int i_unique = 0
    
    center = np.zeros(d, dtype=np.double)
    cdef np.ndarray[double, ndim=1] center_view = center
    
    cdef double hyp_correction
                
    cdef double * dist
        
    cdef int *close_i_hyp, *close_i_hyp_full
    cdef int card_close_i_hyp
    
    close_i_hyp_full = <int *> malloc(n_hyp * sizeof(int *))
    
    cdef int cnt_trigger_hyp
    
    # discretize
    X_digit = discretize(X=X, 
                         x_min = x_min,
                         n = n,
                         d = d,
                         h = h,
                         shift = shift)
    
    # sort
    sort_according_d(X_digit, n, d)
    
    # count
    for j in range(d):
        last_x[j] = X_digit[0][j]
    i_unique = -1
    i_last = 0
    
    for i in range(1,n):
        trigger_unique = False
        for j in range(d):
            if X_digit[i][j] != last_x[j] or (i == n-1 and j == 0):
                # if it's the first column to change
                if trigger_unique == False:
                    # their is a new unique !
                    # increment
                    i_unique = i_unique + 1
                                            
                    # # for each column
                    for j_unique in range(d):
                        # save center for boundary correction below
                        center_view[j_unique] = x_min[j_unique] - 1.5 * h + shift + h * last_x[j_unique]
                        # save the unique x
                        X_digit_uniques_view[i_unique, j_unique] = last_x[j_unique]
                    
                    
                    hyp_correction = 1.0
                    
                    # <!-- Boundary correction
                    dist = inf_distances_to_hyperplanes(center=center_view,
                                                        d=d,
                                                        A = A,
                                                        R = R,
                                                        n_hyp=n_hyp,
                                                        norm_inf_w=norm_inf_w,
                                                        norm_2_w = norm_2_w)
                    
                    card_close_i_hyp = 0
                    for i_hyp in range(n_hyp):
                        if dist[i_hyp] < h/2:
                            close_i_hyp_full[card_close_i_hyp] = i_hyp
                            card_close_i_hyp = card_close_i_hyp + 1
                            
                    close_i_hyp = <int *> malloc(card_close_i_hyp * sizeof(int *))
                    for i_hyp in range(card_close_i_hyp):
                        close_i_hyp[i_hyp] = close_i_hyp_full[i_hyp]
                                        
                    if card_close_i_hyp > 0:
                                                
                        A_x = np.zeros((d, card_close_i_hyp), dtype=np.double)
                        for j_unique in range(d):
                            for i_hyp in range(card_close_i_hyp):
                                A_x[j_unique, i_hyp] = A[j_unique][close_i_hyp[i_hyp]]
                            
                        R_x = np.zeros(card_close_i_hyp, dtype=np.double)
                        for i_hyp in range(card_close_i_hyp):
                            R_x[i_hyp] = R[close_i_hyp[i_hyp]]
                            
                            # translation and scale
                            for j_unique in range(d):
                                R_x[i_hyp] = R_x[i_hyp] + A_x[j_unique, i_hyp] * (center_view[j_unique] - h / 2)
                                A_x[j_unique, i_hyp] = A_x[j_unique, i_hyp] * h

                        hyp_correction = hyperfunc.volume(A = A_x,
                                                          R = R_x,
                                                          check_A = False)
                    
                    # end of boundary correction -->
                    
                    # save the proba
                    P_view[i_unique] = (i - i_last) / (n * hyp_correction)
                    i_last = i
                
                # finally save this new last unique
                last_x[j] = X_digit[i][j]
                # and set trigger True
                # to say that a unique is found
                trigger_unique = True
    
    i_to_keep = P>0
    X_digit_uniques = X_digit_uniques[i_to_keep,:]
    P = P[i_to_keep]
    
    for i in range(n):
        free(X_digit[i])
    free(X_digit)
    free(dist)
    free(close_i_hyp)
    free(close_i_hyp_full)
    free(last_x)
    
    return(X_digit_uniques, P, shift)

@cython.boundscheck(False)  # Deactivate bounds checking.
@cython.wraparound(False)   # Deactivate negative indexing.
@cython.cdivision(True)
def merge_predict_shift(np.ndarray[int, ndim=2] X_digit_uniques,
                        np.ndarray[double, ndim=1] P,
                        np.ndarray[double, ndim=2] Y,
                        np.ndarray[double, ndim=1] x_min,
                        double h,
                        double shift):
        
    cdef Py_ssize_t j
    
    cdef int n_X = X_digit_uniques.shape[0]
    cdef int d = X_digit_uniques.shape[1]
    
    cdef int n_Y = Y.shape[0]
    
    cdef int i_X
    
    cdef bool trigger_set_P, keep_search
    
    # Y_digit is [[y0, y1, 0], [y0, y1, 1], ... [y0, y1, n_y]]
    # by this way, the index is kept after the sort process
    cdef int **Y_digit
    
    
    f = np.zeros(n_Y, dtype=np.double)
    cdef np.ndarray[double, ndim=1] f_view = f
    
    cdef int last_j_success_level, j_success_level
    
    # discretize
    Y_digit = discretize(X = Y,
                     x_min = x_min,
                     n = n_Y,
                     d = d,
                     h = h,
                     shift = shift,
                     index_column = True)
    
    # sort
    sort_according_d(Y_digit, n_Y, d)
    
    i_X = -1
    
    for i_Y in range(n_Y):
        trigger_set_P = False
        
        keep_search = True
        
        while i_X < n_X and keep_search:
            i_X = i_X + 1
            
            for j in range(d):
                if Y_digit[i_Y][j] > X_digit_uniques[i_X, j]:
                    break
                elif Y_digit[i_Y][j] < X_digit_uniques[i_X, j]:
                    keep_search = False
                    break
                else:
                    if keep_search and j + 1 == d:
                        f_view[Y_digit[i_Y][d]] = f_view[Y_digit[i_Y][d]] + P[i_X]
                        keep_search = False
        
        i_X = i_X -1
    
    for i in range(n_Y):
        free(Y_digit[i])
    free(Y_digit)
    
    return(f)

@cython.boundscheck(False)  # Deactivate bounds checking.
@cython.wraparound(False)   # Deactivate negative indexing.
cdef bool lower(int *a,
                int *b,
                int d):
    for j in range(d):
        if a[j] < b[j]:
            return(True)
    return(False)

@cython.boundscheck(False)  # Deactivate bounds checking.
@cython.wraparound(False)   # Deactivate negative indexing.
@cython.cdivision(True)
cdef int ** discretize(np.ndarray[double, ndim=2] X,
                       np.ndarray[double, ndim=1] x_min,
                       int n,
                       int d,
                       double h,
                       double shift,
                       bool index_column = False):
    cdef Py_ssize_t i, j
    
    cdef int **X_digit
    X_digit = <int **> malloc(n * sizeof(int*))
    
    cdef double *slide = <double *> malloc(d * sizeof(double))
    for j in range(d):
        slide[j] = x_min[j] - 2 * h + shift
    
    if index_column:
        for i in range(n):
            X_digit[i] = <int *> malloc((d + 1) * sizeof(int))
            for j in range(d):
                X_digit[i][j] = <int>((X[i,j] - slide[j]) / h)
            
            X_digit[i][d] = i
        
    else:
        for i in range(n):
            X_digit[i] = <int *> malloc(d * sizeof(int))
            for j in range(d):
                X_digit[i][j] = <int>((X[i,j] - slide[j]) / h)
    
    free(slide)
    return(X_digit)

@cython.boundscheck(False)  # Deactivate bounds checking.
@cython.wraparound(False)   # Deactivate negative indexing.
@cython.cdivision(True)
cdef double * inf_distances_to_hyperplanes(np.ndarray[double, ndim=1] center,
                                           int d,
                                           np.ndarray[double, ndim=2] A, 
                                           np.ndarray[double, ndim=1] R,
                                           int n_hyp,
                                           np.ndarray[double, ndim=1] norm_inf_w,
                                           np.ndarray[double, ndim=1] norm_2_w):
    
    cdef double * dist
    dist = <double *> malloc(n_hyp * sizeof(double))
    cdef Py_ssize_t i_hyp, j
    
    # initialize dist_min greater than the limit condition
    for i_hyp in range(n_hyp):
        # distance to hyperplan
        dist[i_hyp] = 0.0
        for j in range(d):
            dist[i_hyp] = dist[i_hyp] + center[j] * A[j, i_hyp]
        dist[i_hyp] = dist[i_hyp] + R[i_hyp]
        # dist[i_hyp] = abs(dist[i_hyp]) / norm_2_w[i_hyp]
        dist[i_hyp] = abs(dist[i_hyp]) / norm_2_w[i_hyp]**2 * norm_inf_w[i_hyp]
        
    return(dist)

@cython.boundscheck(False)  # Deactivate bounds checking.
@cython.wraparound(False)   # Deactivate negative indexing.
@cython.cdivision(True)
cdef double hyperplanes_clip_volume_monte_carlo(np.ndarray[double, ndim=1] center,
                                    int d, 
                                    np.ndarray[double, ndim=2] X_mc,
                                    int n_mc, 
                                    np.ndarray[double, ndim=2] A,
                                    np.ndarray[double, ndim=1] r,
                                    int n_hyp, 
                                    bool * trigger_hyp):
    cdef double dot
    cdef double hyp_correction = 0.0
    cdef bool trigger_mc_inside
    
    cdef Py_ssize_t i_hyp, i_mc, j
    for i_mc in range(n_mc):
        trigger_mc_inside = True
        for i_hyp in range(n_hyp):
            if trigger_hyp[i_hyp]:
                dot = 0.0
                for j in range(d):
                    dot = dot + (center[j]+X_mc[i_mc, j]) * A[j, i_hyp]
                dot = dot + r[i_hyp]
                                                    
                if dot < 0.0:
                    trigger_mc_inside = False
        
        if trigger_mc_inside:
            hyp_correction = hyp_correction + 1
        
    hyp_correction = hyp_correction / n_mc
    
    return(hyp_correction)

cdef void sort_according_d(void *base, 
                           int n,
                           int d):
    
    if d == 1:
        qsort(base, n, sizeof(int*), compare_1d)
    elif d == 2:
        qsort(base, n, sizeof(int*), compare_2d)
    elif d == 3:
        qsort(base, n, sizeof(int*), compare_3d)
    elif d == 4:
        qsort(base, n, sizeof(int*), compare_4d)
    elif d == 5:
        qsort(base, n, sizeof(int*), compare_5d)
    elif d == 6:
        qsort(base, n, sizeof(int*), compare_6d)
    elif d == 7:
        qsort(base, n, sizeof(int*), compare_7d)
    elif d == 8:
        qsort(base, n, sizeof(int*), compare_8d)
    elif d == 9:
        qsort(base, n, sizeof(int*), compare_9d)
    elif d == 10:
        qsort(base, n, sizeof(int*), compare_10d)
    elif d == 11:
        qsort(base, n, sizeof(int*), compare_11d)
    elif d == 12:
        qsort(base, n, sizeof(int*), compare_12d)
    elif d == 13:
        qsort(base, n, sizeof(int*), compare_13d)
    elif d == 14:
        qsort(base, n, sizeof(int*), compare_14d)
    elif d == 15:
        qsort(base, n, sizeof(int*), compare_15d)
    elif d == 16:
        qsort(base, n, sizeof(int*), compare_16d)
    elif d == 17:
        qsort(base, n, sizeof(int*), compare_17d)
    elif d == 18:
        qsort(base, n, sizeof(int*), compare_18d)
    elif d == 19:
        qsort(base, n, sizeof(int*), compare_19d)
    elif d == 20:
        qsort(base, n, sizeof(int*), compare_20d)

# =============================================================================
# Compare functions
# =============================================================================

@cython.boundscheck(False)  # Deactivate bounds checking.
@cython.wraparound(False)   # Deactivate negative indexing.
@cython.cdivision(True) # Deactivate zero division checking.
cdef int compare_1d(const void *a, const void *b):
    cdef Py_ssize_t j
    
    cdef int *ai = (<int**>a)[0]
    cdef int *bi = (<int**>b)[0]
    
    for j in range(1):
        if ai[j] < bi[j]:
            return -1
        if ai[j] > bi[j]:
            return +1
    
    return 0

@cython.boundscheck(False)  # Deactivate bounds checking.
@cython.wraparound(False)   # Deactivate negative indexing.
@cython.cdivision(True) # Deactivate zero division checking.
cdef int compare_2d(const void *a, const void *b):
    cdef Py_ssize_t j
    
    cdef int *ai = (<int**>a)[0]
    cdef int *bi = (<int**>b)[0]
    
    for j in range(2):
        if ai[j] < bi[j]:
            return -1
        if ai[j] > bi[j]:
            return +1
    
    return 0

@cython.boundscheck(False)  # Deactivate bounds checking.
@cython.wraparound(False)   # Deactivate negative indexing.
@cython.cdivision(True) # Deactivate zero division checking.
cdef int compare_3d(const void *a, const void *b):
    cdef Py_ssize_t j
    
    cdef int *ai = (<int**>a)[0]
    cdef int *bi = (<int**>b)[0]
    
    for j in range(3):
        if ai[j] < bi[j]:
            return -1
        if ai[j] > bi[j]:
            return +1
    
    return 0

@cython.boundscheck(False)  # Deactivate bounds checking.
@cython.wraparound(False)   # Deactivate negative indexing.
@cython.cdivision(True) # Deactivate zero division checking.
cdef int compare_4d(const void *a, const void *b):
    cdef Py_ssize_t j
    
    cdef int *ai = (<int**>a)[0]
    cdef int *bi = (<int**>b)[0]
    
    for j in range(4):
        if ai[j] < bi[j]:
            return -1
        if ai[j] > bi[j]:
            return +1
    
    return 0

@cython.boundscheck(False)  # Deactivate bounds checking.
@cython.wraparound(False)   # Deactivate negative indexing.
@cython.cdivision(True) # Deactivate zero division checking.
cdef int compare_5d(const void *a, const void *b):
    cdef Py_ssize_t j
    
    cdef int *ai = (<int**>a)[0]
    cdef int *bi = (<int**>b)[0]
    
    for j in range(5):
        if ai[j] < bi[j]:
            return -1
        if ai[j] > bi[j]:
            return +1
    
    return 0

@cython.boundscheck(False)  # Deactivate bounds checking.
@cython.wraparound(False)   # Deactivate negative indexing.
@cython.cdivision(True) # Deactivate zero division checking.
cdef int compare_6d(const void *a, const void *b):
    cdef Py_ssize_t j
    
    cdef int *ai = (<int**>a)[0]
    cdef int *bi = (<int**>b)[0]
    
    for j in range(6):
        if ai[j] < bi[j]:
            return -1
        if ai[j] > bi[j]:
            return +1
    
    return 0

@cython.boundscheck(False)  # Deactivate bounds checking.
@cython.wraparound(False)   # Deactivate negative indexing.
@cython.cdivision(True) # Deactivate zero division checking.
cdef int compare_7d(const void *a, const void *b):
    cdef Py_ssize_t j
    
    cdef int *ai = (<int**>a)[0]
    cdef int *bi = (<int**>b)[0]
    
    for j in range(7):
        if ai[j] < bi[j]:
            return -1
        if ai[j] > bi[j]:
            return +1
    
    return 0

@cython.boundscheck(False)  # Deactivate bounds checking.
@cython.wraparound(False)   # Deactivate negative indexing.
@cython.cdivision(True) # Deactivate zero division checking.
cdef int compare_8d(const void *a, const void *b):
    cdef Py_ssize_t j
    
    cdef int *ai = (<int**>a)[0]
    cdef int *bi = (<int**>b)[0]
    
    for j in range(8):
        if ai[j] < bi[j]:
            return -1
        if ai[j] > bi[j]:
            return +1
    
    return 0

@cython.boundscheck(False)  # Deactivate bounds checking.
@cython.wraparound(False)   # Deactivate negative indexing.
@cython.cdivision(True) # Deactivate zero division checking.
cdef int compare_9d(const void *a, const void *b):
    cdef Py_ssize_t j
    
    cdef int *ai = (<int**>a)[0]
    cdef int *bi = (<int**>b)[0]
    
    for j in range(9):
        if ai[j] < bi[j]:
            return -1
        if ai[j] > bi[j]:
            return +1
    
    return 0

@cython.boundscheck(False)  # Deactivate bounds checking.
@cython.wraparound(False)   # Deactivate negative indexing.
@cython.cdivision(True) # Deactivate zero division checking.
cdef int compare_10d(const void *a, const void *b):
    cdef Py_ssize_t j
    
    cdef int *ai = (<int**>a)[0]
    cdef int *bi = (<int**>b)[0]
    
    for j in range(10):
        if ai[j] < bi[j]:
            return -1
        if ai[j] > bi[j]:
            return +1
    
    return 0

@cython.boundscheck(False)  # Deactivate bounds checking.
@cython.wraparound(False)   # Deactivate negative indexing.
@cython.cdivision(True) # Deactivate zero division checking.
cdef int compare_11d(const void *a, const void *b):
    cdef Py_ssize_t j
    
    cdef int *ai = (<int**>a)[0]
    cdef int *bi = (<int**>b)[0]
    
    for j in range(11):
        if ai[j] < bi[j]:
            return -1
        if ai[j] > bi[j]:
            return +1
    
    return 0

@cython.boundscheck(False)  # Deactivate bounds checking.
@cython.wraparound(False)   # Deactivate negative indexing.
@cython.cdivision(True) # Deactivate zero division checking.
cdef int compare_12d(const void *a, const void *b):
    cdef Py_ssize_t j
    
    cdef int *ai = (<int**>a)[0]
    cdef int *bi = (<int**>b)[0]
    
    for j in range(12):
        if ai[j] < bi[j]:
            return -1
        if ai[j] > bi[j]:
            return +1
    
    return 0

@cython.boundscheck(False)  # Deactivate bounds checking.
@cython.wraparound(False)   # Deactivate negative indexing.
@cython.cdivision(True) # Deactivate zero division checking.
cdef int compare_13d(const void *a, const void *b):
    cdef Py_ssize_t j
    
    cdef int *ai = (<int**>a)[0]
    cdef int *bi = (<int**>b)[0]
    
    for j in range(13):
        if ai[j] < bi[j]:
            return -1
        if ai[j] > bi[j]:
            return +1
    
    return 0

@cython.boundscheck(False)  # Deactivate bounds checking.
@cython.wraparound(False)   # Deactivate negative indexing.
@cython.cdivision(True) # Deactivate zero division checking.
cdef int compare_14d(const void *a, const void *b):
    cdef Py_ssize_t j
    
    cdef int *ai = (<int**>a)[0]
    cdef int *bi = (<int**>b)[0]
    
    for j in range(14):
        if ai[j] < bi[j]:
            return -1
        if ai[j] > bi[j]:
            return +1
    
    return 0

@cython.boundscheck(False)  # Deactivate bounds checking.
@cython.wraparound(False)   # Deactivate negative indexing.
@cython.cdivision(True) # Deactivate zero division checking.
cdef int compare_15d(const void *a, const void *b):
    cdef Py_ssize_t j
    
    cdef int *ai = (<int**>a)[0]
    cdef int *bi = (<int**>b)[0]
    
    for j in range(15):
        if ai[j] < bi[j]:
            return -1
        if ai[j] > bi[j]:
            return +1
    
    return 0

@cython.boundscheck(False)  # Deactivate bounds checking.
@cython.wraparound(False)   # Deactivate negative indexing.
@cython.cdivision(True) # Deactivate zero division checking.
cdef int compare_16d(const void *a, const void *b):
    cdef Py_ssize_t j
    
    cdef int *ai = (<int**>a)[0]
    cdef int *bi = (<int**>b)[0]
    
    for j in range(16):
        if ai[j] < bi[j]:
            return -1
        if ai[j] > bi[j]:
            return +1
    
    return 0

@cython.boundscheck(False)  # Deactivate bounds checking.
@cython.wraparound(False)   # Deactivate negative indexing.
@cython.cdivision(True) # Deactivate zero division checking.
cdef int compare_17d(const void *a, const void *b):
    cdef Py_ssize_t j
    
    cdef int *ai = (<int**>a)[0]
    cdef int *bi = (<int**>b)[0]
    
    for j in range(17):
        if ai[j] < bi[j]:
            return -1
        if ai[j] > bi[j]:
            return +1
    
    return 0

@cython.boundscheck(False)  # Deactivate bounds checking.
@cython.wraparound(False)   # Deactivate negative indexing.
@cython.cdivision(True) # Deactivate zero division checking.
cdef int compare_18d(const void *a, const void *b):
    cdef Py_ssize_t j
    
    cdef int *ai = (<int**>a)[0]
    cdef int *bi = (<int**>b)[0]
    
    for j in range(18):
        if ai[j] < bi[j]:
            return -1
        if ai[j] > bi[j]:
            return +1
    
    return 0

@cython.boundscheck(False)  # Deactivate bounds checking.
@cython.wraparound(False)   # Deactivate negative indexing.
@cython.cdivision(True) # Deactivate zero division checking.
cdef int compare_19d(const void *a, const void *b):
    cdef Py_ssize_t j
    
    cdef int *ai = (<int**>a)[0]
    cdef int *bi = (<int**>b)[0]
    
    for j in range(19):
        if ai[j] < bi[j]:
            return -1
        if ai[j] > bi[j]:
            return +1
    
    return 0

@cython.boundscheck(False)  # Deactivate bounds checking.
@cython.wraparound(False)   # Deactivate negative indexing.
@cython.cdivision(True) # Deactivate zero division checking.
cdef int compare_20d(const void *a, const void *b):
    cdef Py_ssize_t j
    
    cdef int *ai = (<int**>a)[0]
    cdef int *bi = (<int**>b)[0]
    
    for j in range(20):
        if ai[j] < bi[j]:
            return -1
        if ai[j] > bi[j]:
            return +1
    
    return 0



#!/usr/bin/env python3

# Set this to True to enable building extensions using Cython.
# Set it to False to build extensions from the C file (that
# was previously created using Cython).
# Set it to 'auto' to build with Cython if available, otherwise
# from the C file.
USE_CYTHON = 'auto'

import sys

import numpy

from setuptools import setup
from setuptools import Extension

__version__ = "0.1.2"

if USE_CYTHON:
    try:
        from Cython.Distutils import build_ext
    except ImportError:
        if USE_CYTHON=='auto':
            USE_CYTHON=False
        else:
            raise

cmdclass = { }
ext_modules = [ ]

if USE_CYTHON:
    ext_modules += [
        Extension("fastash.ashfunc",
                  sources=["cython/ashfunc.pyx"],
                  include_dirs=[numpy.get_include()],
                  language="c++",
                  ),
    ]
    cmdclass.update({ 'build_ext': build_ext })
else:
    ext_modules += [
        Extension("fastash.ashfunc", 
                  sources=["cython/ashfunc.cpp"],
                  include_dirs=[numpy.get_include()],
                  language="c++"),
    ]

setup(
    name='fastash',
    version=__version__,
    description='Fast Averages Shifted Histogram',
    author='François-Rémi Mazy',
    author_email='francois-remi.mazy@inria.fr',
    url='https://gitlab.inria.fr/fmazy/fastash',
    packages=[ 'fastash', ],
    package_dir={
        'fastash' : 'fastash',
    },
    cmdclass = cmdclass,
    ext_modules=ext_modules,
    install_requires=[
            'numpy>=1.19.2',
            'hyperclip>=0.1.2',
        ],

    long_description=open('README.md').read(),
    long_description_content_type = "text/markdown",
    license="BSD 3-Clause License",
    classifiers=[
        'Development Status :: 1 - Planning',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: BSD License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Cython',
        'Topic :: Scientific/Engineering :: Mathematics',
    ],
    keywords='averaged shifted histograms kde ash density estimation kernel',
)

